@@ -1,5 +1,5 @@
 /*
-** Copyright (C) 2010 Sourcefire, Inc.
+** Copyright (C) 2010-2013 Sourcefire, Inc.
 ** Author: Michael R. Altizer <maltizer@sourcefire.com>
 **
 ** This program is free software; you can redistribute it and/or modify
@@ -15,7 +15,7 @@
 **
 ** You should have received a copy of the GNU General Public License
 ** along with this program; if not, write to the Free Software
-** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+** Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
 
 #ifndef _DAQ_COMMON_H
@@ -79,21 +79,73 @@
 #define DAQ_ERROR_EXISTS    -8  /* Argument or device already exists */
 #define DAQ_READFILE_EOF    -42 /* Hit the end of the file being read! */
 
-#define DAQ_PKT_FLAG_HW_TCP_CS_GOOD  0x1
+#define DAQ_PKT_FLAG_HW_TCP_CS_GOOD     0x1 /* The DAQ module reports that the checksum for this packet is good. */
+#define DAQ_PKT_FLAG_OPAQUE_IS_VALID    0x2 /* The DAQ module actively set the opaque value in the DAQ packet header. */
+#define DAQ_PKT_FLAG_NOT_FORWARDING     0x4 /* The DAQ module will not be actively forwarding this packet
+                                               regardless of the verdict (e.g, Passive or Inline Tap interfaces). */
+#define DAQ_PKT_FLAG_PRE_ROUTING        0x8 /* The packet is being routed via us but packet modifications
+                                                (MAC and TTL) have not yet been made. */
 
-/* The DAQ packet header structure passed to DAQ Analysis Functions.  This should NEVER be modified by user applications. */
+/* The DAQ packet header structure passed to DAQ Analysis Functions.
+ * This should NEVER be modified by user applications. */
+#define DAQ_PKTHDR_UNKNOWN  -1  /* Ingress or Egress not known */
+#define DAQ_PKTHDR_FLOOD    -2  /* Egress is flooding */
 typedef struct _daq_pkthdr
 {
     struct timeval ts;      /* Timestamp */
     uint32_t caplen;        /* Length of the portion present */
     uint32_t pktlen;        /* Length of this packet (off wire) */
-    int device_index;       /* Index of the receiving interface. */
+    int32_t ingress_index;  /* Index of the inbound interface. */
+    int32_t egress_index;   /* Index of the outbound interface. */
+    int32_t ingress_group;  /* Index of the inbound group. */
+    int32_t egress_group;   /* Index of the outbound group. */
     uint32_t flags;         /* Flags for the packet (DAQ_PKT_FLAG_*) */
+    uint32_t opaque;        /* Opaque context value from the DAQ module or underlying hardware.
+                               Directly related to the opaque value in FlowStats. */
+    void *priv_ptr;         /* Private data pointer */
+    uint16_t address_space_id; /* Unique ID of the address space */
 } DAQ_PktHdr_t;
 
+#define DAQ_METAHDR_TYPE_SOF    0
+#define DAQ_METAHDR_TYPE_EOF    1
+typedef struct _daq_metahdr
+{
+    int type;               /* Type */
+} DAQ_MetaHdr_t;
+
+typedef struct _daq_modflow
+{
+    uint32_t    opaque;     /* */
+} DAQ_ModFlow_t;
+
+#define DAQ_FLOWSTATS_IPV4      0
+#define DAQ_FLOWSTATS_IPV6      1
+typedef struct _flow_stats
+{
+    int32_t ingressZone;
+    int32_t egressZone;
+    int32_t ingressIntf;
+    int32_t egressIntf;
+    uint8_t initiatorIp[16];
+    uint8_t responderIp[16];
+    uint16_t initiatorPort;
+    uint16_t responderPort;
+    uint8_t protocol;
+    uint8_t version;
+    uint64_t initiatorPkts;
+    uint64_t responderPkts;
+    uint64_t initiatorBytes;
+    uint64_t responderBytes;
+    uint32_t opaque;
+    uint16_t vlan_tag;
+    struct timeval sof_timestamp;
+    struct timeval eof_timestamp;
+    uint16_t address_space_id;
+} Flow_Stats_t, *Flow_Stats_p;
+
 typedef enum {
     DAQ_VERDICT_PASS,       /* Pass the packet. */
-    DAQ_VERDICT_BLOCK,       /* Block the packet. */
+    DAQ_VERDICT_BLOCK,      /* Block the packet. */
     DAQ_VERDICT_REPLACE,    /* Pass a packet that has been modified in-place. (No resizing allowed!) */
     DAQ_VERDICT_WHITELIST,  /* Pass the packet and fastpath all future packets in the same flow systemwide. */
     DAQ_VERDICT_BLACKLIST,  /* Block the packet and block all future packets in the same flow systemwide. */
@@ -102,6 +154,7 @@ typedef enum {
 } DAQ_Verdict;
 
 typedef DAQ_Verdict (*DAQ_Analysis_Func_t)(void *user, const DAQ_PktHdr_t *hdr, const uint8_t *data);
+typedef int (*DAQ_Meta_Func_t)(void *user, const DAQ_MetaHdr_t *hdr, const uint8_t *data);
 
 typedef enum {
     DAQ_MODE_PASSIVE,
@@ -136,12 +189,12 @@ typedef enum {
 
 typedef struct _daq_stats
 {
-    uint64_t hw_packets_received;           /* Packets received by the hardware */
-    uint64_t hw_packets_dropped;            /* Packets dropped by the hardware */
-    uint64_t packets_received;              /* Packets received by this instance */
-    uint64_t packets_filtered;              /* Packets filtered by this instance's BPF */
-    uint64_t packets_injected;              /* Packets injected by this instance */
-    uint64_t verdicts[MAX_DAQ_VERDICT];     /* Counters of packets handled per-verdict. */
+    uint64_t hw_packets_received;       /* Packets received by the hardware */
+    uint64_t hw_packets_dropped;        /* Packets dropped by the hardware */
+    uint64_t packets_received;          /* Packets received by this instance */
+    uint64_t packets_filtered;          /* Packets filtered by this instance's BPF */
+    uint64_t packets_injected;          /* Packets injected by this instance */
+    uint64_t verdicts[MAX_DAQ_VERDICT]; /* Counters of packets handled per-verdict. */
 } DAQ_Stats_t;
 
 /* DAQ module type flags */
