@@ -310,6 +310,8 @@ static int ipfw_daq_acquire (
 
         if ( select(impl->sock+1, &fdset, NULL, NULL, &tv) < 0 )
         {
+            if ( errno == EINTR )
+                break;
             DPE(impl->error, "%s: can't select divert socket (%s)\n",
                 __FUNCTION__, strerror(errno));
             return DAQ_ERROR;
